﻿section Impero_API;

PageLimit = 10000;

[DataSource.Kind="Impero_API", Publish="Impero_API.Publish"]
shared Impero_API.Contents = (tenant as text) =>
    let
        // Control activities data  
        #"Fetched Control Activities" = Impero_API.FetchControlActivities(tenant),
  
        // Control activities resources
        #"Fetched Control Activities Resources" = Impero_API.FetchControlActivitiesResources(tenant),

        // Prepare the Control Program dimension
        #"Expanded Control Programs from Control Activities Resources" = Table.ExpandRecordColumn(#"Fetched Control Activities Resources", "Column1", {"controlPrograms"}, {"Control Programs"}),
        #"Expanded Control Programs" = Table.ExpandListColumn(#"Expanded Control Programs from Control Activities Resources", "Control Programs"),
        #"Expanded Control Program Fields" = Table.ExpandRecordColumn(#"Expanded Control Programs", "Control Programs", {"id", "name"}, {"Control Program ID", "Control Program"}),
        #"Cleaned Control Programs" = Table.Distinct(#"Expanded Control Program Fields"),
     
        // Prepare the Control dimension
        #"Expanded Controls from Control Activities Resources" = Table.ExpandRecordColumn(#"Fetched Control Activities Resources", "Column1", {"controls"}, {"Control"}),
        #"Expanded Controls" = Table.ExpandListColumn(#"Expanded Controls from Control Activities Resources", "Control"),
        #"Expanded Control Fields" = Table.ExpandRecordColumn(#"Expanded Controls", "Control", {"id", "title", "description"}, {"Control ID", "Control", "Control Description"}),
        #"Cleaned Controls" = Table.Distinct(#"Expanded Control Fields"),
     
        // Prepare the User dimension
        #"Expanded Users from Control Activities Resources" = Table.ExpandRecordColumn(#"Fetched Control Activities Resources", "Column1", {"users"}, {"User"}),
        #"Expanded Users" = Table.ExpandListColumn(#"Expanded Users from Control Activities Resources", "User"),
        #"Expanded User Fields" = Table.ExpandRecordColumn(#"Expanded Users", "User", {"id", "email", "lastname", "firstname"}, {"User ID", "Email", "Last name", "First name"}),
        #"Cleaned Users" = Table.Distinct(#"Expanded User Fields"),

        // Prepare Control Activities fact
        #"Expanded Control Activities" = Table.ExpandListColumn(#"Fetched Control Activities", "Column1"),
        #"Expanded Control Activities Record" = Table.ExpandRecordColumn(#"Expanded Control Activities", "Column1", 
        {"id", "controlProgramId", "title", "description", "completed", "financialYear", "period", "assignments", "tags", "archived", "loginRequired", "segregationOfDutiesEnforced", "createdAt", "controlId", "tasks"}, 
        {"Control Activity ID", "Control Program ID", "Control Title", "Control Description", "Completed", "Financial Year", "Period", "Assignments", "Tags", "Archived", "Login Required", "Segregation Of Duties Enforced", "Created At", "Control ID", "Control Tasks"}),
        #"Cleaned Control Activities" = Table.TransformColumnTypes(#"Expanded Control Activities Record",{{"Archived", type logical}, {"Login Required", type logical}, {"Segregation Of Duties Enforced", type logical}, {"Created At", type datetime}})
        ,
        // Navigation table exposed to the user of the custom connector
        objects = #table(
            {"Name", "Key", "Data", "ItemKind", "ItemName", "IsLeaf"}, {
            {"Control Activities", "data", #"Cleaned Control Activities", "Table", "Table", true},
            {"Control", "controls", #"Cleaned Controls", "Table", "Table", true},
            {"Control Program", "controlPrograms", #"Cleaned Control Programs", "Table", "Table", true},
            {"User", "users", #"Cleaned Users", "Table", "Table", true}
        }),
        #"Navigation Table" = Table.ToNavigationTable(objects, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        #"Navigation Table";

// Data Source Kind description
Impero_API = [
    TestConnection = (dataSourcePath) =>
        let
            json = Json.Document(dataSourcePath),
            tenant = json[tenant]
        in
            { "Impero_API.Contents", tenant},
    Authentication = [
        Key = []
    ],
    Label = Extension.LoadString("DataSourceLabel")
];

// Data Source UI publishing description
Impero_API.Publish = [
    Beta = true,
    Category = "Other",
    ButtonText = { Extension.LoadString("ButtonTitle"), Extension.LoadString("ButtonHelp") },
    LearnMoreUrl = "https://support.impero.com/",
    SourceImage = Impero_API.Icons,
    SourceTypeImage = Impero_API.Icons
];

Impero_API.Icons = [
    Icon16 = { Extension.Contents("Impero_API16.png"), Extension.Contents("Impero_API20.png"), Extension.Contents("Impero_API24.png"), Extension.Contents("Impero_API32.png") },
    Icon32 = { Extension.Contents("Impero_API32.png"), Extension.Contents("Impero_API40.png"), Extension.Contents("Impero_API48.png"), Extension.Contents("Impero_API64.png") }
];


// Impero API fetching code

shared Impero_API.FetchControlActivities = (tenant as text) =>
    let
        ApiKey = Extension.CurrentCredential()[Key],
        Paging = List.Generate( () =>
            [
                Url = "https://" & tenant & "/public-api/v1/control-activities?limit=" & Number.ToText(PageLimit),
                WebContents = Web.Contents(Url, [Headers=[#"X-Api-Key"=ApiKey]]),
                Headers =  Value.Metadata(WebContents),
                HttpHeaders = Headers[Headers],
                NextPage = ""
            ],
            each [NextPage] <> null,
            each [
                NextPage = try Record.Field([HttpHeaders], "X-Pagination-Next") otherwise null,
                Url = "https://" & tenant & "/public-api/v1/control-activities?limit=" & Number.ToText(PageLimit) & "&page=" & NextPage,
                WebContents = Web.Contents(Url, [Headers=[#"X-Api-Key"=ApiKey]]),
                Headers =  Value.Metadata(WebContents),
                HttpHeaders = Headers[Headers]
            ],
            each Record.Field(Json.Document([WebContents]), "data")),
        #"Converted Paging to Table" = Table.FromList(Paging, Splitter.SplitByNothing(), null, null, ExtraValues.Error)
   in
        #"Converted Paging to Table";

shared Impero_API.FetchControlActivitiesResources = (tenant as text) =>
    let
        ApiKey = Extension.CurrentCredential()[Key],
        Paging = List.Generate( () =>
            [
                Url = "https://" & tenant & "/public-api/v1/control-activities?limit=" & Number.ToText(PageLimit),
                WebContents = Web.Contents(Url, [Headers=[#"X-Api-Key"=ApiKey]]),
                Headers =  Value.Metadata(WebContents),
                HttpHeaders = Headers[Headers],
                NextPage = ""
            ],
            each [NextPage] <> null,
            each [
                NextPage = try Record.Field(HttpHeaders, "X-Pagination-Next") otherwise null,
                Url = "https://" & tenant & "/public-api/v1/control-activities?limit=" & Number.ToText(PageLimit) & "&page=" & NextPage,
                WebContents = Web.Contents(Url, [Headers=[#"X-Api-Key"=ApiKey]]),
                Headers =  Value.Metadata(WebContents),
                HttpHeaders = Headers[Headers]
            ],
            each Record.Field(Json.Document([WebContents]), "resources")),
        #"Converted Paging to Table" = Table.FromList(Paging, Splitter.SplitByNothing(), null, null, ExtraValues.Error)
   in
        #"Converted Paging to Table";


// Common library code

Table.ToNavigationTable = (
    table as table,
    keyColumns as list,
    nameColumn as text,
    dataColumn as text,
    itemKindColumn as text,
    itemNameColumn as text,
    isLeafColumn as text
) as table =>
    let
        tableType = Value.Type(table),
        newTableType = Type.AddTableKey(tableType, keyColumns, true) meta 
        [
            NavigationTable.NameColumn = nameColumn, 
            NavigationTable.DataColumn = dataColumn,
            NavigationTable.ItemKindColumn = itemKindColumn, 
            Preview.DelayColumn = itemNameColumn, 
            NavigationTable.IsLeafColumn = isLeafColumn
        ],
        navigationTable = Value.ReplaceType(table, newTableType)
    in
        navigationTable;